/*
 * DSP ReplayGain plugin for LAMIP-0.0.3
 *
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <plug_dsp.h>
#include <plug_control.h> // to read from the songinfo in DSPMaterial


static int	lamip_dsprg_init(const lamipPluginHandle*);
static void	lamip_dsprg_cleanup(void);
static int	lamip_dsprg_setup(lamipPCMInfo*);
static int	lamip_dsprg_probe(DSPMaterial*);
static int	lamip_dsprg_filter(DSPMaterial*);

static DSPPlugin dsprg;

static float	scale = 0;
static int	use_album = 1;
static int	anticlip = 1;

#ifdef DEBUGING
 #define DEBUG(a) fprintf(stderr, "%s\n", a);
#else
 #define DEBUG(a)
#endif

/****************************************************************************/

DSPPlugin* lamip_dsp_info(void)
{
  dsprg.common.name		= "replaygain";
  dsprg.common.description	= "replaygain - reads ReplayGain infos from tags and modifies the stream with them";
  dsprg.common.init		= lamip_dsprg_init;
  dsprg.common.cleanup		= lamip_dsprg_cleanup;
  dsprg.setup			= lamip_dsprg_setup;
  dsprg.probe			= lamip_dsprg_probe;
  dsprg.filter			= lamip_dsprg_filter;

  return(&dsprg);
}

static int	lamip_dsprg_init(const lamipPluginHandle* conf)
{
  lamipPluginHandle*	config = (lamipPluginHandle *)conf;
  int			val_use_album;
  int			val_use_anticlip;

  DEBUG("replaygain init");
  if(!lamip_cfg_getExist(conf, "use_album")){
    lamip_send_message("REPLAY_GAIN: lamip_dsprg_init: no config value for \"use_album\" found, resetting to default...\n");
    val_use_album = 1;
    lamip_cfg_setBool(config, "use_album", val_use_album);
  }else{
    val_use_album = lamip_cfg_getBool(conf, "use_album");
  }
  use_album = val_use_album;
  if(!lamip_cfg_getExist(conf, "use_anticlip")){
    lamip_send_message("REPLAY_GAIN: lamip_dsprg_init: no config value for \"use_anticlip\" found, resetting to default...\n");
    val_use_anticlip = 1;
    lamip_cfg_setBool(config, "use_anticlip", val_use_anticlip);
  }else{
    val_use_anticlip = lamip_cfg_getBool(conf, "use_anticlip");
  }
  anticlip = val_use_anticlip;
  return(1);
}

static void	lamip_dsprg_cleanup(void)
{
  DEBUG("replaygain_cleanup");
}

static int lamip_dsprg_setup(lamipPCMInfo* setup)
{
  /* We don't change any samplerate or format or channel... so just
   * return */
  DEBUG("replaygain_setup");
  return(1);
}

static int	lamip_dsprg_probe(DSPMaterial* material)
{
  lamipSonginfo* si = material->songinfo;	
  char*	tmp	= 0;
  float	peak	= 0;


  /* As we don't modify any lamipPCMInfo things... just give the output
   * the input information, but check if it's PCM_FORMAT_S16_* format
   * other things aren't handled at the moment */
  DEBUG("replaygain_probe");
#ifdef WORDS_BIGENDIAN
  if(material->input_pcm_info.format != PCM_FORMAT_S16_BE) {
#else
  if(material->input_pcm_info.format != PCM_FORMAT_S16_LE) {
#endif
    lamip_send_message("REPLAY_GAIN: Cannot apply ReplayGain to other format then S16! Aborting...\n");
    return(0);
  }
  material->output_pcm_info = material->input_pcm_info;
  /* this is a good place to get the ReplayGain infos */
  if(use_album){
    tmp = lamip_songinfo_getField(si, "REPLAYGAIN_ALBUM_GAIN");
    if(!tmp)
      tmp = lamip_songinfo_getField(si, "RG_AUDIOPHILE"); /* obsolete name */
  }
  if(!tmp){
    use_album = 0;
    tmp = lamip_songinfo_getField(si, "REPLAYGAIN_TRACK_GAIN");
  }
  if(!tmp)
    tmp = lamip_songinfo_getField(si, "RG_RADIO"); /* obsolete name */
  if(tmp){
    scale = pow(10., atof(tmp) / 20);
    free(tmp);
  }else{
    scale = 1.;
  }
  tmp = 0;
  if(anticlip) {
    if(use_album)
      tmp = lamip_songinfo_getField(si, "REPLAYGAIN_ALBUM_PEAK");
    if(!tmp)
      tmp = lamip_songinfo_getField(si, "REPLAYGAIN_TRACK_PEAK");
    if(!tmp)
      tmp = lamip_songinfo_getField(si, "RG_PEAK"); /* obsolete name */
    if(tmp){
      peak = atof(tmp);
      free(tmp);
    }else{
      peak = 1.;
    }
    if(scale*peak > 1.)
      scale = 1. / peak;
  }
#ifdef DEBUGING
  lamip_send_message("REPLAY_GAIN: lamip_dsprg_probe: scaling with %f!\n", scale);
#endif
  return(1);
}

static int lamip_dsprg_filter(DSPMaterial* material)
{
  int i;
  int tmp;
  short* in;
  short* out;

  /* Allocating the output buffer first */
  material->output_buffer_len = material->input_buffer_len;
  material->output_buffer = malloc(material->output_buffer_len);
  in = material->input_buffer;
  out = material->output_buffer;
  for(i=0; i<(material->input_buffer_len/2); i++){
    tmp = in[i] * scale;
    if(tmp > 0x7fff)
      tmp = 0x7fff;
    else if(tmp < -0x8000)
      tmp = -0x8000;
    out[i] = tmp;
  }
  return(1);
}
